@@ -29,8 +29,6 @@
 #include <libart_lgpl/art_rgb_rgba_affine.h>
 #include "gnome-canvas-pixbuf.h"
 
-
-
 /* Private part of the GnomeCanvasPixbuf structure */
 typedef struct {
 	/* Our gdk-pixbuf */
@@ -66,8 +64,6 @@ typedef struct {
 	GtkAnchorType anchor;
 } PixbufPrivate;
 
-
-
 /* Object argument IDs */
 enum {
 	PROP_0,
@@ -123,25 +119,27 @@ static GnomeCanvasItemClass *parent_class;
 GtkType
 gnome_canvas_pixbuf_get_type (void)
 {
-	static GtkType canvas_pixbuf_type = 0;
+	static GType pixbuf_type;
 
-	if (!canvas_pixbuf_type) {
-		static const GtkTypeInfo canvas_pixbuf_info = {
-			"GnomeCanvasPixbuf",
-			sizeof (GnomeCanvasPixbuf),
+	if (!pixbuf_type) {
+		static const GTypeInfo object_info = {
 			sizeof (GnomeCanvasPixbufClass),
-			(GtkClassInitFunc) gnome_canvas_pixbuf_class_init,
-			(GtkObjectInitFunc) gnome_canvas_pixbuf_init,
-			NULL, /* reserved_1 */
-			NULL, /* reserved_2 */
-			(GtkClassInitFunc) NULL
+			(GBaseInitFunc) NULL,
+			(GBaseFinalizeFunc) NULL,
+			(GClassInitFunc) gnome_canvas_pixbuf_class_init,
+			(GClassFinalizeFunc) NULL,
+			NULL,			/* class_data */
+			sizeof (GnomeCanvasPixbuf),
+			0,			/* n_preallocs */
+			(GInstanceInitFunc) gnome_canvas_pixbuf_init,
+			NULL			/* value_table */
 		};
 
-		canvas_pixbuf_type = gtk_type_unique (gnome_canvas_item_get_type (),
-						      &canvas_pixbuf_info);
+		pixbuf_type = g_type_register_static (GNOME_TYPE_CANVAS_ITEM, "GnomeCanvasPixbuf",
+						     &object_info, 0);
 	}
 
-	return canvas_pixbuf_type;
+	return pixbuf_type;
 }
 
 /* Class initialization function for the pixbuf canvas item */
@@ -156,7 +154,7 @@ gnome_canvas_pixbuf_class_init (GnomeCanvasPixbufClass *class)
 	object_class = (GtkObjectClass *) class;
 	item_class = (GnomeCanvasItemClass *) class;
 
-	parent_class = gtk_type_class (gnome_canvas_item_get_type ());
+	parent_class = g_type_class_peek_parent (class);
 
 	gobject_class->set_property = gnome_canvas_pixbuf_set_property;
 	gobject_class->get_property = gnome_canvas_pixbuf_get_property;
@@ -257,6 +255,7 @@ gnome_canvas_pixbuf_init (GnomeCanvasPixbuf *gcp)
 	priv->height = 0.0;
 	priv->x = 0.0;
 	priv->y = 0.0;
+	priv->anchor = GTK_ANCHOR_NW;
 }
 
 /* Destroy handler for the pixbuf canvas item */
@@ -596,7 +595,7 @@ compute_viewport_affine (GnomeCanvasPixbuf *gcp, double *viewport_affine, double
 	case GTK_ANCHOR_NW:
 	case GTK_ANCHOR_W:
 	case GTK_ANCHOR_SW:
-		ti_len += x;
+		ti_len *= x;
 		break;
 
 	case GTK_ANCHOR_N:
@@ -659,12 +658,18 @@ compute_viewport_affine (GnomeCanvasPixbuf *gcp, double *viewport_affine, double
  * by some other amount.
  */
 static void
-compute_render_affine (GnomeCanvasPixbuf *gcp, double *render_affine, double *i2c)
+compute_render_affine (GnomeCanvasPixbuf *gcp, double *ra, double *i2c)
 {
-	double viewport_affine[6];
+	double va[6];
 
-	compute_viewport_affine (gcp, viewport_affine, i2c);
-	art_affine_multiply (render_affine, viewport_affine, i2c);
+	compute_viewport_affine (gcp, va, i2c);
+#ifdef GNOME_CANVAS_PIXBUF_VERBOSE
+	g_print ("va %g %g %g %g %g %g\n", va[0], va[1], va[2], va[3], va[4], va[5]);
+#endif
+	art_affine_multiply (ra, va, i2c);
+#ifdef GNOME_CANVAS_PIXBUF_VERBOSE
+	g_print ("ra %g %g %g %g %g %g\n", ra[0], ra[1], ra[2], ra[3], ra[4], ra[5]);
+#endif
 }
 
 /* Recomputes the bounding box of a pixbuf canvas item.  The horizontal and
@@ -672,11 +677,11 @@ compute_render_affine (GnomeCanvasPixbuf *gcp, double *render_affine, double *i2
  * have to compute the components individually for each dimension.
  */
 static void
-recompute_bounding_box (GnomeCanvasPixbuf *gcp)
+recompute_bounding_box (GnomeCanvasPixbuf *gcp, gdouble *i2c)
 {
 	GnomeCanvasItem *item;
 	PixbufPrivate *priv;
-	double i2c[6], render_affine[6];
+	double ra[6];
 	ArtDRect rect;
 
 	item = GNOME_CANVAS_ITEM (gcp);
@@ -693,9 +698,18 @@ recompute_bounding_box (GnomeCanvasPixbuf *gcp)
 	rect.y0 = 0.0;
 	rect.y1 = gdk_pixbuf_get_height (priv->pixbuf);
 
+#ifdef GNOME_CANVAS_PIXBUF_VERBOSE
+	g_print ("i2c %g %g %g %g %g %g\n", i2c[0], i2c[1], i2c[2], i2c[3], i2c[4], i2c[5]);
+#endif
 	gnome_canvas_item_i2c_affine (item, i2c);
-	compute_render_affine (gcp, render_affine, i2c);
-	art_drect_affine_transform (&rect, &rect, render_affine);
+#ifdef GNOME_CANVAS_PIXBUF_VERBOSE
+	g_print ("i2c %g %g %g %g %g %g\n", i2c[0], i2c[1], i2c[2], i2c[3], i2c[4], i2c[5]);
+#endif
+	compute_render_affine (gcp, ra, i2c);
+#ifdef GNOME_CANVAS_PIXBUF_VERBOSE
+	g_print ("ra %g %g %g %g %g %g\n", ra[0], ra[1], ra[2], ra[3], ra[4], ra[5]);
+#endif
+	art_drect_affine_transform (&rect, &rect, ra);
 
 	item->x1 = floor (rect.x0);
 	item->y1 = floor (rect.y0);
@@ -720,12 +734,19 @@ gnome_canvas_pixbuf_update (GnomeCanvasItem *item, double *affine, ArtSVP *clip_
 	if (parent_class->update)
 		(* parent_class->update) (item, affine, clip_path, flags);
 
+        /* the optimzations below cause rarely triggered redrawing bugs and
+	 * don't seem to actually save much performance. so it's probably
+	 * better to turn them off, than to chase subtle optimization bugs
+	 * throughgout all of gnome-canvas-pixbuf.c - TIMJ
+	 */
+#if USE_BROKEN_OPTIMIZATIONS
 	if (((flags & GNOME_CANVAS_UPDATE_VISIBILITY)
 	     && !(GTK_OBJECT_FLAGS (item) & GNOME_CANVAS_ITEM_VISIBLE))
 	    || (flags & GNOME_CANVAS_UPDATE_AFFINE)
 	    || priv->need_pixbuf_update
-	    || priv->need_xform_update)
+	    || priv->need_xform_update) {
 		gnome_canvas_request_redraw (item->canvas, item->x1, item->y1, item->x2, item->y2);
+	}
 
 	/* If we need a pixbuf update, or if the item changed visibility to
 	 * shown, recompute the bounding box.
@@ -735,11 +756,21 @@ gnome_canvas_pixbuf_update (GnomeCanvasItem *item, double *affine, ArtSVP *clip_
 	    || ((flags & GNOME_CANVAS_UPDATE_VISIBILITY)
 		&& (GTK_OBJECT_FLAGS (gcp) & GNOME_CANVAS_ITEM_VISIBLE))
 	    || (flags & GNOME_CANVAS_UPDATE_AFFINE)) {
-		recompute_bounding_box (gcp);
+		recompute_bounding_box (gcp, affine);
+#ifdef GNOME_CANVAS_PIXBUF_VERBOSE
+		g_print ("BBox is %g %g %g %g\n", item->x1, item->y1, item->x2, item->y2);
+#endif
 		gnome_canvas_request_redraw (item->canvas, item->x1, item->y1, item->x2, item->y2);
 		priv->need_pixbuf_update = FALSE;
 		priv->need_xform_update = FALSE;
 	}
+#else   /* ordinary update logic */
+        gnome_canvas_request_redraw (item->canvas, item->x1, item->y1, item->x2, item->y2);
+        recompute_bounding_box (gcp, affine);
+        gnome_canvas_request_redraw (item->canvas, item->x1, item->y1, item->x2, item->y2);
+        priv->need_pixbuf_update = FALSE;
+        priv->need_xform_update = FALSE;
+#endif
 }
 
 
@@ -857,8 +888,8 @@ gnome_canvas_pixbuf_draw (GnomeCanvasItem *item, GdkDrawable *drawable,
 					     0, 0,
 					     d_rect.x0 - x, d_rect.y0 - y,
 					     w, h,
-					     GDK_PIXBUF_ALPHA_BILEVEL,
-					     128,
+					     GDK_PIXBUF_ALPHA_FULL,
+					     0,
 					     GDK_RGB_DITHER_MAX,
 					     d_rect.x0, d_rect.y0);
 
@@ -884,7 +915,55 @@ gnome_canvas_pixbuf_render (GnomeCanvasItem *item, GnomeCanvasBuf *buf)
 	compute_render_affine (gcp, render_affine, i2c);
         gnome_canvas_buf_ensure_buf (buf);
 
-	if (gdk_pixbuf_get_has_alpha(priv->pixbuf))
+
+	if ((fabs (render_affine[1]) < GNOME_CANVAS_EPSILON) &&
+	    (fabs (render_affine[2]) < GNOME_CANVAS_EPSILON) &&
+	    render_affine[0] > 0.0 &&
+	    render_affine[3] > 0.0)
+	  {
+	    GdkPixbuf *dest_pixbuf;
+	    int x0, y0, x1, y1;
+
+	    dest_pixbuf = gdk_pixbuf_new_from_data (buf->buf,
+						    GDK_COLORSPACE_RGB,
+						    FALSE,
+						    8,
+						    buf->rect.x1 - buf->rect.x0,
+						    buf->rect.y1 - buf->rect.y0,
+						    buf->buf_rowstride,
+						    NULL, NULL);
+
+
+	    x0 = floor (render_affine[4] - buf->rect.x0 + 0.5);
+	    y0 = floor (render_affine[5] - buf->rect.y0 + 0.5);
+
+	    x1 = x0 + floor (gdk_pixbuf_get_width (priv->pixbuf) * render_affine[0] + 0.5);
+	    y1 = y0 + floor (gdk_pixbuf_get_height (priv->pixbuf) * render_affine[3] + 0.5);
+
+	    x0 = MAX (x0, 0);
+	    x0 = MIN (x0, buf->rect.x1 - buf->rect.x0);
+	    y0 = MAX (y0, 0);
+	    y0 = MIN (y0, buf->rect.y1 - buf->rect.y0);
+	    
+	    x1 = MAX (x1, 0);
+	    x1 = MIN (x1, buf->rect.x1 - buf->rect.x0);
+	    y1 = MAX (y1, 0);
+	    y1 = MIN (y1, buf->rect.y1 - buf->rect.y0);
+	    
+	    gdk_pixbuf_composite  (priv->pixbuf,
+				   dest_pixbuf,
+				   x0, y0,
+				   x1 - x0, y1 - y0,
+				   render_affine[4] - buf->rect.x0,
+				   render_affine[5] - buf->rect.y0,
+				   render_affine[0],
+				   render_affine[3],
+				   GDK_INTERP_NEAREST,
+				   255);
+
+	    gdk_pixbuf_unref (dest_pixbuf);
+	  }
+	else if (gdk_pixbuf_get_has_alpha(priv->pixbuf))
 		art_rgb_rgba_affine (buf->buf,
 				     buf->rect.x0, buf->rect.y0, buf->rect.x1, buf->rect.y1,
 				     buf->buf_rowstride,
